;;;;--------------------------------------------
;;;; An interface to talk to breve simulations -
;;;;                                           -
;;;; Currently supported protocols are:        -
;;;; - HTTP over TCP/IP.                       -
;;;;                                           -
;;;; Requirements of breve simulation:         -
;;;; - Must respond to the method "get-iteration-times". The method should return "0" when 
;;;;   breve has finished computations.
;;;;--------------------------------------------


;;;---------------------
;;; Package definition -
;;;---------------------
(defpackage #:breve-interface
  (:use "COMMON-LISP")
  (:export
    #:execute-breve-command
    #:execute-breve-command-with-result
    #:init-breve
    #:wait-for-breve))

(in-package :breve-interface)


;;;--------------------
;;; Global parameters -
;;;--------------------
(defparameter *breve-host* "localhost")
(defparameter *breve-port* 8000)
(defparameter *protocol* 'HTTP/1.0_TCP/IP)
(defparameter *debug* nil)


;;;------------
;;; Functions -
;;;------------
(defun init-breve ()
  "Start a breve server"
  #+Lispworks(sys:call-system "start-breve-server"))

(eval-when (:compile-toplevel :execute :load-toplevel)
  #+SBCL(require :sb-bsd-sockets))

(defun wait-for-breve ()
  "Wait for simulation to have 0 integrations left to do."
  (loop until (= (read-from-string (execute-breve-command-with-result
				    "get-integration-times"))
		 0)
       do
       #+win32 (sleep 0.5)
       #-win32(sleep 0.05)))
     
(defun execute-breve-command (command &rest arguments)
  "Execute a breve command with it's arguments."
  (case *protocol*
    ('HTTP/1.0_TCP/IP
     (with-open-stream 
	 (http
	  #+Lispworks (comm:open-tcp-stream *breve-host* *breve-port*)
	  #+OpenMCL (openmcl-socket:make-socket
		     :address-family :internet
		     :type :stream 
		     :remote-host *breve-host*
		     :remote-port *breve-port*)
	  #+SBCL (let ((s (sb-bsd-sockets:make-inet-socket :stream :tcp)))
		   (sb-bsd-sockets:socket-connect
		    s (sb-bsd-sockets:get-host-by-name *breve-host*))
		   s)
	  )
       ;; Send the command to breve using GET on the commands URL
       (let ((command (format nil "GET /~A~A HTTP/1.0~C~C~C~C"
                              command
                              (loop for a in arguments
                                    collect (format nil "_~D" a) into list
                                    finally (return (apply 'concatenate 'string list)))
                              (code-char 13) (code-char 10) (code-char 13) (code-char 10))))
         (when *debug*
           (format t "~A" command))
         (format http "~A" command))
       (force-output http)))))

(defun execute-breve-command-with-result (command &rest arguments)
  "Execute a breve command with it's arguments, and return the result from breve"
  (case *protocol*
    ('HTTP/1.0_TCP/IP
     (with-open-stream
	 (http 
	  #+Lispworks (comm:open-tcp-stream *breve-host* *breve-port*)
	  #+OpenMCL (openmcl-socket:make-socket
		     :address-family :internet
		     :type :stream 
		     :remote-host *breve-host* 
		     :remote-port *breve-port*))
       ;; Send the command to breve using GET on the commands URL
       (let ((command (format nil "GET /~A~A HTTP/1.0~C~C~C~C"
                              command
                              (loop for a in arguments
                                    collect (format nil "_~D" a) into list
                                    finally (return (apply 'concatenate 'string list)))
                              (code-char 13) (code-char 10) (code-char 13) (code-char 10))))
         (when *debug*
           (format t "~A" command))
         (format http "~A" command))
       (force-output http)
       ;;;;     ;; Wait for a reply to arrive
       (loop for char = (read-char-no-hang http nil :eof)
             until char
             do
             #+win32 (sleep 0.1) ;; Breves' HTTP is a bit slow on windows, for some reason.
             #-win32(sleep 0.01) ;; Wait a millisecond before retrying 
             finally (unless (eq char :eof) (unread-char char http)))
       ;;     ;; Now, collect the reply and return it
       (read-line http nil nil)))))