;;;; Applying the Genetic Algorithm to the Beer-part of the leg controller
;;;;
;;;; What to evolve:
;;;;
;;;; 0 1:
;;;; - Stance and swing neurons (all are equal):
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;;
;;;; 2 3:
;;;; - Foot neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep* 10]
;;;;
;;;; 4 5 6:
;;;; - Sensor neurons (all are equal):
;;;;   - External current magnitude [0 255]
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep 10] 
;;;;
;;;; 7 8:
;;;; - Pacemaker neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:timestep* 10]
;;;;
;;;; Genotype representation is a cons on the form:
;;;;   ('vector of unsigned bytes' . fitness)
;;;;
;;;;----------------------------------------------------------------------------

(in-package :carausius-morosus)


;;;;-----------------------------------------
;;;; Utility functions, unrelated to the GA -
;;;;-----------------------------------------
(defun make-random-number-between (min max &optional (decimals 0))
  "Create a random number between min and max (inclusive)."
  (let ((max (* (expt 10 decimals) max))
        (min (* (expt 10 decimals) min)))
    (/ (+ min (random (- max min -1))) (expt 10 decimals))))

(defun get-location ()
  "Get the location of the breve creature."
  (car
   (read-from-string
    (coerce
     (loop for char across
           (breve-interface:execute-breve-command-with-result "get-location")
           unless (char-equal #\, char)
           collect char) 'string))))

;;;;---------------
;;;; GA functions -
;;;;---------------
(defun make-random-population (size)
  "Create a list of length 'size' of random genotypes."
  (loop for n from 1 to size
        collect (let ((a (make-array 9 :element-type '(unsigned-byte 17)))) ; about 16,8 mil. values
                  (loop for n from 0 to 8
                        do (setf (aref a n) (make-random-number-between 0 (expt 2 24))))
                  (cons a 0)))) ; Individual initialized neither fit nor unfit

(defun genotype->phenotype (genotype)
  "Convert a genotype into a creature that is suitable for fitness evaluation."
  (let* ((cns (make-instance 'central-nervous-system))
         (genome (car genotype))
         (locomotion-controller
          (central-nervous-system-locomotion-controller cns))
         (leg-controllers
          (list (locomotion-controller-l1 locomotion-controller)
                (locomotion-controller-l2 locomotion-controller)
                (locomotion-controller-l3 locomotion-controller)
                (locomotion-controller-r1 locomotion-controller)
                (locomotion-controller-r2 locomotion-controller)
                (locomotion-controller-r3 locomotion-controller))))
    ;; NOTE: Inhibit the LC-neuron for now
    (setf (neuron-external-current
           (locomotion-controller-lc-neuron
            (central-nervous-system-locomotion-controller cns)))
          -100)
    (dolist (l leg-controllers)
      ;; Set pacemaker units to be single pacemaker neurons (see appendix)
      (setf (pacemaker-io-neuron (leg-controller-pacemaker l))
            (make-instance 'ctrnn::pacemaker-neuron))
      (setf (ctrnn::pacemaker-neuron-intrinsic-low-firing-time-slope
             (pacemaker-io-neuron (leg-controller-pacemaker l)))
            -356.2)
      (setf (ctrnn::pacemaker-neuron-intrinsic-low-firing-time-intercept
             (pacemaker-io-neuron (leg-controller-pacemaker l)))
             675.0)
      (setf (ctrnn::pacemaker-neuron-intrinsic-high-firing-time
             (pacemaker-io-neuron (leg-controller-pacemaker l)))
            372.0)
      ;; Reconnect pacemakers to the rest of the leg
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (leg-controller-lc-neuron l)
                      :strength 10))
      (add-synapse!
       (leg-controller-stance-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron
                                    (leg-controller-pacemaker l))
                      :strength -10))
      (add-synapse! 
       (leg-controller-swing-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron
                                    (leg-controller-pacemaker l))
                      :strength 10))
      (add-synapse!
       (leg-controller-foot-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron
                                    (leg-controller-pacemaker l))
                      :strength -4))
      (add-synapse! 
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (leg-controller-backward-angle-sensor l)
                      :strength 10))
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (leg-controller-forward-angle-sensor l)
                      :strength -15))
      ;; Set the evolved bias/time for the pacemakers
      (setf (neuron-bias (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (+ (* (- (aref genome 7) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (- (* (+ (aref genome 8) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; [0 16] + timestep 
      ;; Set stance and swing neurons bias/time
      ;;---------------------------------------
      (setf (neuron-bias (leg-controller-stance-neuron l))
            (+ (* (- (aref genome 0) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (leg-controller-stance-neuron l))
            (- (* (+ (aref genome 1) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; [0 16] + timestep 
      (setf (neuron-bias (leg-controller-swing-neuron l))
            (+ (* (- (aref genome 0) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (leg-controller-swing-neuron l))
            (- (* (+ (aref genome 1) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; [0 16] + timestep 
      ;; Foot neuron bias/time
      ;;----------------------
      (setf (neuron-bias (leg-controller-foot-neuron l))
            (+ (* (- (aref genome 2) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (leg-controller-foot-neuron l))
            (- (* (+ (aref genome 3) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; [0 16] + timestep 
      ;; Sensor neurons (forward back)
      ;;------------------------------
      (setf (neuron-external-current (leg-controller-forward-angle-sensor l))
            (+ (* (- (aref genome 4) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-bias (leg-controller-forward-angle-sensor l))
            (+ (* (- (aref genome 5) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (leg-controller-forward-angle-sensor l))
            (- (* (+ (aref genome 6) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; [0 16] + timestep 
      (setf (neuron-external-current (leg-controller-backward-angle-sensor l))
            (+ (* (- (aref genome 4) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-bias (leg-controller-backward-angle-sensor l))
            (+ (* (- (aref genome 5) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
      (setf (neuron-time-constant (leg-controller-backward-angle-sensor l))
            (- (* (+ (aref genome 6) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*))) ; [0 16] + timestep
    cns))

(defun fitness (genotype)
  "Evaluate the fitness of a genotype, and return the fitness."
  (breve-interface:execute-breve-command "reset")
  (let ((cns (genotype->phenotype genotype))
        (energy 0)
        (start-location (get-location)))
    (dotimes (var 200)
      (dotimes (var 5)
        (incf energy
              (synchronously-update-membrane-potentials!
               (central-nervous-system-locomotion-controller cns))))
      (synchronize-with-breve (central-nervous-system-breve-object cns)))
    (let ((distance-walked (- (get-location) start-location)))
      (/ distance-walked energy))))
  
(defun evaluate (individuals)
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        collect (progn
                  (format t "~&evaluating ~A ...~%" (car genotype))
                  (let ((res (cons (car genotype)
                                   (fitness genotype))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with one-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (x (make-random-number-between 1 (- (length c1) 1))))
    (setf (subseq c1 0 x) (subseq (car p2) 0 x))
    (setf (subseq c2 0 x) (subseq (car p1) 0 x))
    (list (cons c1 0) (cons c2 0))))


(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (random 2)))
    (cond ((= 1 rand)
           (when (< (aref new-genome x) (- (expt 2 24) 100))
             (incf (aref new-genome x) (random 100))))
          (t
           (when (> (aref new-genome x) 100)
             (decf (aref new-genome x) (random 100)))))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations)
  (breve-interface:execute-breve-command "set-output-filter" -1)
  (let ((plot-pane (make-instance 'capi-extensions:plot-pane
                                  :x-min -10 :x-max num-generations
                                  :x-axis (/ num-generations 10)
                                  :y-min -10.0
                                  :y-max 40.0
                                  :y-axis 10))
        (best-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
        (avg-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :yellow))
        (worst-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :red))
        (population (sort-population (evaluate (make-random-population 10)))))
    (push best-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push avg-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push worst-graph (capi-extensions:plot-pane-graphs plot-pane))
    (capi:contain plot-pane)
    (do ((generation 1 (+ generation 1)))
        ((> generation num-generations))
      (format t "*** Generation ~A~%" generation)
      ;; Select parents
      (let ((parents (select-parents population 4)))
        ;; Recombine
        (let ((children-1 (recombine (list (first parents) (second parents))))
              (children-2 (recombine (list (third parents) (fourth parents)))))
          ;; Maybe mutate
          (when (< (random 100) 5)
            (setf children-1 (loop for c in children-1
                                 collect (mutate c))))
          (when (< (random 100) 5)
            (setf children-2 (loop for c in children-2
                                 collect (mutate c))))
          ;; Evaluate new individuals
          (setf children-1 (evaluate children-1))
          (setf children-2 (evaluate children-2))
          ;; Select individuals for next generation
          (setf population (sort-population (append (butlast population 4)
                                                    children-1 children-2)))
          ;; Update display
          (push (cons generation (cdr (first population)))
                (capi-extensions::plot-pane-graph-points best-graph))
          (push (cons generation (/ (loop for ind in population
                                          sum (cdr ind))
                                    (length population)))
                (capi-extensions::plot-pane-graph-points avg-graph))
          (push (cons generation (cdr (first (last population))))
                (capi-extensions::plot-pane-graph-points worst-graph))
          (gp:invalidate-rectangle plot-pane))))
    (first population)))