;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Applying the Genetic Algorithm to the Beer-part of the leg controller
;;;;
;;;; * What is given:
;;;;
;;;; 0 1 2 3:
;;;; - Pacemaker unit neurons (different for each neuron):
;;;;   - Bias [-20 20]/[-20 0]
;;;;   - Time constant [ctrnn:*timestep* 10]
;;;;
;;;; 4 5 6 7:
;;;; - Weights between pacemaker unit neurons:
;;;;   - P1 -> P1 [4 20]
;;;;   - P1 -> P2 [-20 20]
;;;;   - P2 -> P1 [4 20]
;;;;   - P2 -> P2 [-20 20]
;;;;
;;;; * What to evolve:
;;;;
;;;; 8 9:
;;;; - Stance neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 16]
;;;;
;;;; 10 11:
;;;; - Swing neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 16]
;;;; 
;;;; 12 13:
;;;; - Foot neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep* 16]
;;;;
;;;; 14 15:
;;;; - Sensor neurons (all are equal):
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep 16] 
;;;;
;;;; 16 17 18 19 20 21 22 23 24: 
;;;; - Weights between neurons:
;;;;   - pacemaker->stance [-20 0]
;;;;   - pacemaker->swing [0 20]
;;;;   - pacemaker->foot [-20 0]
;;;;   - backwd.s.->pacemaker [0 20]
;;;;   - forwrd.s.->pacemaker [-20 0]
;;;;   - forwrd.s.->stance [0 20]
;;;;   - forwrd.s.->swing [-20 0]
;;;;   - forwrd.s.->foot [0 20]
;;;;   - between all pacemakers [-20 0]
;;;;
;;;; Genotype representation is a cons on the form:
;;;;   ('vector of unsigned bytes' . fitness)
;;;;
;;;; This take corresponds to the creature in carausisu-morosus-6 !
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(in-package :leg-controller)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))
(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun weight-gene->weight (gene)
  "Convert 24-bit integer into a real weight value."
  (+ -20 (/ gene (/ (expt 2 24) 40))))

(defun time-constant-gene->time-constant (gene)
  (+ (/ gene (/ (expt 2 24) 16)) ctrnn::*timestep*))

(defun weight->weight-gene (weight)
  "Convert real weight value into 24-bit integer."
  (round (* (+ weight 20) (/ (expt 2 24) 40))))

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; GA functions
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-population (size)
  "Create a list of length 'size' of random genotypes."
  (loop for n from 1 to size
        collect (let ((a (make-array
			  25 :element-type '(unsigned-byte 24)))) ; ~16,8m vals
		  (setf (subseq a 0 8) ; Replace by pacemaker genome
			(vector ; Taken from best of pacemaker runs (see log-8)
			 10960798 524456 2452260 526275
			 11386161 251453 12616268 16028625))
		  (loop for n from 0 to 24
                        do (cond (;; Randomize pacemaker genome to 'neigborhood'
				  ;; of that evolved in pacemaker evolution
				  (< n 8) 
				  (setf (aref a n)
					(incf (aref a n)
					      (- (random 20000)
						 10000))))
				 ;; Randomize other genome to values close to
				 ;; values that correspond to the controller
				 ;; of beer-1990
		;		 ((= n 8)
		;		  (setf (aref a n)
		;			8388608)) ; stance neuron bias
                ;                 ((= n 10)
                ;                  (setf (aref a n)
		;			9437184)) ; swing neuron bias
                ;                 ((= n 12)
                ;                  (setf (aref a n)
		;			(+ 7600000 
		;			   (- 100000 
		;			      (make-random-number-between
		;			       0 200000))))) ; foot neuron bias
                ;                 ((= n 14)
                ;                  (setf (aref a n)
		;			8388608)) ; sensory neurons bias
				 ((= n 16) ; pacemaker->stance ~-10
                                  (let ((middle (/ (expt 2 24) 4)))
                                    (setf (aref a n)
					  (make-random-number-between
					   (round (- middle 1000000))
					   (round (+ middle 1000000))))))
                                 ((or (= n 17) ; pacemaker->swing ~10
				      (= n 19) ; bw.ang.->pacemaker ~10
				      (= n 21) ; fw.ang.->stance ~10
				      (= n 23)) ; fw.ang.->foot ~10
                                  (let ((middle (- (expt 2 24)
						   (/ (expt 2 24) 4))))
                                    (setf (aref a n)
					  (make-random-number-between
					   (round (- middle 1000000))
					   (round (+ middle 1000000))))))
                                 ((= n 18) ; pacemaker->foot ~4
				  (let ((quart (/ (expt 2 24) 20/8)))
				    (setf (aref a n)
					  (make-random-number-between 
					   (round (- quart 1000000))
					   (round (+ quart 1000000))))))
                                 ((or (= n 20) ; fw.ang.->pacemaker ~15 
				      (= n 22) ; fw.ang.->swing ~15
				      (= n 24)) ; pacemakers ~15
				  (let ((quart (/ (expt 2 24) 8)))
                                    (setf (aref a n)
					  (make-random-number-between 
					   (round (- quart 1000000))
					   (round (+ quart 1000000))))))
                                 (t
                                  (setf (aref a n) 
					(make-random-number-between 
					 0 (expt 2 24))))))
                  (cons a 0)))) ; Individual initialized neither fit nor unfit

(defun genotype->phenotype (genotype)
  "Convert a genotype into a creature that is suitable for fitness evaluation."
  (let* ((cns (make-instance 'central-nervous-system))
         (genome (car genotype))
         (locomotion-controller
          (central-nervous-system-locomotion-controller cns))
         (leg-controllers
          (list (locomotion-controller-l1 locomotion-controller)
                (locomotion-controller-l2 locomotion-controller)
                (locomotion-controller-l3 locomotion-controller)
                (locomotion-controller-r1 locomotion-controller)
                (locomotion-controller-r2 locomotion-controller)
                (locomotion-controller-r3 locomotion-controller))))
    (dolist (l leg-controllers)
      ;; Pacemaker (gene 0-7)
      ;;---------------------
      ;; Set pacemaker neuron variables from the genome
      (setf (neuron-bias 
	     (pacemaker-io-neuron (leg-controller-pacemaker l)))
	    (weight-gene->weight (aref genome 0)))
      (setf (neuron-time-constant
	     (pacemaker-io-neuron (leg-controller-pacemaker l)))
	    (time-constant-gene->time-constant (aref genome 1)))
      (setf (neuron-bias
	     (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
	    (weight-gene->weight (aref genome 2)))
      (setf (neuron-time-constant
	     (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
	    (time-constant-gene->time-constant (aref genome 3)))
      ;; Create synapses within the pacemaker circuit
      (setf (neuron-dendrites
	     (pacemaker-io-neuron 
	      (leg-controller-pacemaker l))) nil) ; Remove existing
      ;; IO neuron synapses:
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
		      :from-neuron (pacemaker-io-neuron
				    (leg-controller-pacemaker l))
		      :strength (weight-gene->weight (aref genome 4))))
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
		      :from-neuron (pacemaker-hidden-neuron 
				    (leg-controller-pacemaker l))
		      :strength (weight-gene->weight (aref genome 5))))
      ;; Hidden neuron synapses (same procedure):
      (setf (neuron-dendrites
	     (pacemaker-hidden-neuron (leg-controller-pacemaker l))) nil)
      (add-synapse!
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse 
		      :from-neuron (pacemaker-hidden-neuron
				    (leg-controller-pacemaker l))
		      :strength (weight-gene->weight (aref genome 6))))
      (add-synapse!
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
		      :from-neuron (pacemaker-io-neuron
				    (leg-controller-pacemaker l))
		      :strength (weight-gene->weight (aref genome 7))))
      ;; Locomotion controller (gene 8-24)
      ;;----------------------------------
      ;; Set stance and swing neurons bias/time
      (setf (neuron-bias (leg-controller-stance-neuron l))
	    (weight-gene->weight (aref genome 8)))
      (setf (neuron-time-constant (leg-controller-stance-neuron l))
	    (time-constant-gene->time-constant (aref genome 9)))
      (setf (neuron-bias (leg-controller-swing-neuron l))
	    (weight-gene->weight (aref genome 10)))
      (setf (neuron-time-constant (leg-controller-swing-neuron l))
	      (time-constant-gene->time-constant (aref genome 11)))
      ;; Foot neuron bias/time
      (setf (neuron-bias (leg-controller-foot-neuron l))
	    (weight-gene->weight (aref genome 12)))
      (setf (neuron-time-constant (leg-controller-foot-neuron l))
	    (time-constant-gene->time-constant (aref genome 13)))     
      ;; Sensor neurons (forward back)
      (setf (neuron-bias (leg-controller-forward-angle-sensor l))
	    (weight-gene->weight (aref genome 14)))
      (setf (neuron-time-constant (leg-controller-forward-angle-sensor l))
	    (time-constant-gene->time-constant (aref genome 15)))
      (setf (neuron-bias (leg-controller-backward-angle-sensor l))
	    (weight-gene->weight (aref genome 14)))
      (setf (neuron-time-constant (leg-controller-backward-angle-sensor l))
	    (time-constant-gene->time-constant (aref genome 15)))
      ;; Weights
      (add-synapse!
       (leg-controller-stance-neuron l)
       (make-instance 
	'synapse
        :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
	:strength (weight-gene->weight (aref genome 16))))
      (add-synapse!
       (leg-controller-swing-neuron l)
       (make-instance
	'synapse
        :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
	:strength (weight-gene->weight (aref genome 17))))
      (add-synapse!
       (leg-controller-foot-neuron l)
       (make-instance
	'synapse
        :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
	:strength (weight-gene->weight (aref genome 18))))
      (add-synapse! ; Note connects directly to inside of pacemaker
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 
	'synapse
	:from-neuron (leg-controller-backward-angle-sensor l)
	:strength (weight-gene->weight (aref genome 19))))
      (add-synapse! ; Note connects directly to inside of pacemaker
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 
	'synapse
	:from-neuron (leg-controller-forward-angle-sensor l)
	:strength (weight-gene->weight (aref genome 20))))
      (add-synapse!
       (leg-controller-stance-neuron l)
       (make-instance
	'synapse
	:from-neuron (leg-controller-forward-angle-sensor l)
	:strength (weight-gene->weight (aref genome 21))))
      (add-synapse!
       (leg-controller-swing-neuron l)
       (make-instance
	'synapse
	:from-neuron (leg-controller-forward-angle-sensor l)
	:strength (weight-gene->weight (aref genome 22))))
      (add-synapse!
       (leg-controller-foot-neuron l)
       (make-instance
	'synapse
	:from-neuron (leg-controller-forward-angle-sensor l)
	:strength (weight-gene->weight (aref genome 23)))))
    (carausius-morosus::set-up-connections
     locomotion-controller
     (weight-gene->weight (aref genome 24)))
    cns))

(defun fitness (genotype &optional (num-breve-synchs 200))
  "Evaluate the fitness of a genotype, and return the fitness."
  ;; Check if genotype has already been evaluated earlier
  (unless (gethash (car genotype) *fitnesses*)
    (setf (gethash (car genotype) *fitnesses*)
          (progn 
            (breve-interface:execute-breve-command "reset")
            (loop
             until (= 0 (read-from-string
			 (breve-interface:execute-breve-command-with-result
			  "get-integration-times")))
             do (sleep 0.1))
            (let ((cns (genotype->phenotype genotype))
                  (energy 0)
                  (start-location (get-location))
                  (start-time
                   (read-from-string
                    (breve-interface:execute-breve-command-with-result
		     "get-time"))))
              (dotimes (var num-breve-synchs)
                (dotimes (var 5)
                  (incf energy
                        (synchronously-update-membrane-potentials!
                         (central-nervous-system-locomotion-controller cns))))
                (synchronize-with-breve (central-nervous-system-breve-object cns))
                (loop ; Wait for Breve simulation to finish computations
                 until (let
			   ((times
			     (read-from-string
			      (breve-interface:execute-breve-command-with-result
			       "get-integration-times"))))
                         (= 0 times))
		   do (sleep 0.03)))
              (format
	       t " time discrepancy = ~A ~%" ; Debug: Breve & Lisp r synched?
                      (- (* num-breve-synchs 5 ctrnn::*timestep*)
                         (- (read-from-string
                             (breve-interface:execute-breve-command-with-result 
                              "get-time"))
                            start-time)))
              (let ((distance-walked (- (get-location) start-location)))
                (/ distance-walked 1))))))
  (gethash (car genotype) *fitnesses*))
  
(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
                                   (fitness genotype num-breve-synchs))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
  (setf  *fitnesses* (make-hash-table :test #'equalp))
  (breve-interface:execute-breve-command "set-output-filter" -1)
  (let* ((carausius-morosus::*maximum-leg-force* 200000)
         (carausius-morosus::*maximum-leg-down-force* 80000) 
	 ;; NOTE: 70000 = "minimum" to stand on three legs
         #+Lispworks(plot-pane (make-instance 'capi-extensions:plot-pane
                                  :x-min -2 :x-max num-generations 
				  :x-axis (/ num-generations 10)
                                  :y-min -10 ;-0.000000001
                                  :y-max 200 ; 0.00000002
                                  :y-axis 10)); 0.000000002))
        #+Lispworks(best-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
        #+Lispworks(avg-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :yellow))
        #+Lispworks(worst-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :red))
        (population 
	 (sort-population (evaluate (make-random-population 100) 500))))
    #+Lispworks(push best-graph (capi-extensions:plot-pane-graphs plot-pane))
    #+Lispworks(push avg-graph (capi-extensions:plot-pane-graphs plot-pane))
    #+Lispworks(push worst-graph (capi-extensions:plot-pane-graphs plot-pane))
    #+Lispworks(capi:contain plot-pane)
    (with-open-file (f (format nil "logs/leg-take-1-without-cc-log~A.txt" id) :direction :output :if-exists :supersede)
      (do ((generation 1 (+ generation 1)))
          ((> generation num-generations))
        (format t "*** Generation ~A~%" generation)
        ;; Select parents
        (let ((parents (select-parents population 2)))
          ;; Recombine
          (let ((children (recombine parents)))
            ;; Maybe mutate
            (when (< (random 100) 5)
              (setf children (loop for c in children collect (mutate c))))
            ;; Evaluate new individuals
          (setf children (evaluate children 500))
          ;; Select individuals for next generation
          (setf population
                (sort-population (append
                                  children
                                  (remove-individuals population 2))))
          ;; Update display
          (format f "~A ~A ~A ~A ~%"
                  generation
                  (cdr (first population))
                  (/ (loop for ind in population
					 sum (cdr ind))
                     (length population))
                  (cdr (first (last population))))
          (force-output f)
          #+Lispworks
          (progn 
	    (push (cons generation (cdr (first population)))
		  (capi-extensions::plot-pane-graph-points best-graph))
	    (push (cons generation (/ (loop for ind in population
					 sum (cdr ind))
				      (length population)))
		  (capi-extensions::plot-pane-graph-points avg-graph))
	    (push (cons generation (cdr (first (last population))))
		  (capi-extensions::plot-pane-graph-points worst-graph))
	    (gp:invalidate-rectangle plot-pane))))))
    (first population)))

(defun 15-runs ()
  (let (;(breve-interface::*BREVE-HOST* "tsstud02.stud.ntnu.no"))
        ;(breve-interface::*breve-port* 8001)
	)
    (dotimes (var 15)
      (when (> var 8)
	(ga 250 var)))))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (let (;(breve-interface::*BREVE-HOST* "tsstud02.stud.ntnu.no"))
        (breve-interface::*breve-port* 8000))
    (dotimes (var 10)
      (ga 150 var))))


(defun test-run-good ()
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (fitness
     (cons
       (vector 10963896 516852 2460398 475135 11385571 260912 12572387 16024549 8426675 16714521 9437184 1445525 7527271 493102 8388608 5077806 3802889 12303830 6529969 13024977 2308303 13138285 1497664 12821677 1233568) ; best increm.
      0)
     500)))
      