;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Applying SANE-ESP to the complete controller
;;;;
;;;; The genotype is now made up from chromosomes. There's one chromosome per
;;;; neuron, and it is encoded as a list with the following elements:
;;;;
;;;; - Neuron bias
;;;; - Neuron time constant
;;;; - A list of synapes encoded as follows for each synapse:
;;;;   - A number pointing to a neuron (e.g. n'th neuron in the leg-controller)
;;;;   - A number representing the strength of the synapse
;;;;
;;;; Additionally a cummulative fitness is attached, and a count of how many
;;;; times the neuron has been evaluated is attached too.
;;;;
;;;; The population is a list of these individuals. Since ESP is used, the
;;;; population is partitioned into as many partitions as there are neurons
;;;; to be evolved.
;;;;
;;;; The partitions are as follows:
;;;;
;;;; 1.) Front femur-tibia neurons
;;;; 2.) Middle femur-tibia neurons
;;;; 3.) Back femur-tibia neurons

(in-package :leg-controller-take-2)

(shadowing-import 'leg-controller::make-random-number-between)
(shadowing-import 'leg-controller::get-location)

(setf carausius-morosus::*AEP-L1* (+ (* 1.3 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.3 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.5 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.5 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 1.9 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 1.9 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 3.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 3.4 (/ pi 4)) 0.01))

(defstruct chromosome
  (bias-gene (make-random-number-between 0 (expt 2 24)) :type (unsigned-byte 24))
  (time-constant-gene (make-random-number-between 0 (expt 2 24)) :type (unsigned-byte 24))
  (dendrites-genes nil :type list) ; List of cons-"genes", one per synaptic connection (neuron-num . weight)
  (cumulative-fitness 0.0 :type float)
  (num-fitness-evaluations 0 :type integer))

(defun chromosome-average-fitness (chromosome)
  (if (= (chromosome-num-fitness-evaluations chromosome) 0)
    0
    (/ (chromosome-cumulative-fitness chromosome)
       (chromosome-num-fitness-evaluations chromosome))))

(defmethod print-object ((c chromosome)(s stream))
  (format s "B=~,6F, T=~,6F, CF=~,6F, NF=~,D, AF=~,6F, D:~A~%"
          (+ (* (- (chromosome-bias-gene c) (expt 2 24)) (/ 40 (expt 2 24))) 20)
          (- (* (+ (chromosome-time-constant-gene c) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)
          (chromosome-cumulative-fitness c)(chromosome-num-fitness-evaluations c)
          (chromosome-average-fitness c)
          (loop for s in (chromosome-dendrites-genes c)
                collect (cons (car s)
                              (+ (* (- (cdr s) (expt 2 24)) (/ 40 (expt 2 24))) 20.0)))))


(defstruct genome
  (front-leg-chromosome (make-chromosome) :type chromosome)
  (middle-leg-chromosome (make-chromosome) :type chromosome)
  (back-leg-chromosome (make-chromosome) :type chromosome))

(defun make-population (size network-size)
  "Create a random population of neuons."
  (loop for n from 1 to size
        collect (let ((new-chromosome (make-chromosome
                                       :dendrites-genes
                                       (loop for gene from 0 to (random 2) ; NOTE: limit to two connections
                                             collect (cons
                                                      (random network-size)
                                                      (make-random-number-between 0 (expt 2 24)))))))
                  ;; Set the bias-gene so that search will be centered around the most sensitive region of the neuron
                  (let* ((weight-sum
                          (loop for synapse in (chromosome-dendrites-genes new-chromosome)
                                sum  (+ (* (- (cdr synapse) (expt 2 24)) (/ 40 (expt 2 24))) 20)))
                         (bias (/ (- weight-sum) 2)))
                    (setf (chromosome-bias-gene new-chromosome)
                          (round ; Must be represented by unsiged-byte
                           (+ (/ (- (/ (- weight-sum) 2) 20) (/ 40 (expt 2 24))) (expt 2 24))))
                    new-chromosome))))

(defun adjust-to-chromosome! (neuron chromosome network)
  "Adjust 'neuron' to values given by 'chromosome', and attach it to other neurons in 'network'"
  (setf (neuron-bias neuron)
        (+ (* (- (chromosome-bias-gene chromosome) (expt 2 24)) (/ 40 (expt 2 24))) 20)) ; [-20 20]
  (setf (neuron-time-constant neuron)
        (- (* (+ (chromosome-time-constant-gene chromosome) (expt 2 24)) (/ 16 (expt 2 24))) 16 ctrnn::*timestep*)) ; *t.s.* to 16
  (loop for dendrite-gene in (chromosome-dendrites-genes chromosome)
        do (when (nth (car dendrite-gene) (neural-network-neurons network)) ; if such a neuron exists
             (add-synapse! neuron
                           (make-instance 'synapse
                                          :from-neuron (nth (car dendrite-gene) (neural-network-neurons network))
                                          :strength (+ (* (- (cdr dendrite-gene) (expt 2 24)) (/ 40 (expt 2 24))) 20)))))) ; [-20 20]

(defun genotype->phenotype (genome)
  "Create a new cns consisting of a bunch of neurons."
  (let* ((cns ; Creata a cns from the eariler evolved genes
          (leg-controller::genotype->phenotype
           (cons
            (vector
             8741825 343682 3759221 2539167 6218767 2694810 7392821 16234160 ; pacemaker genes
             8388608 4296478 9437184 12538736 7677032 4273179 8388608 12252335 
             8425781 8318696 4267765 8366795 15138802 8382783 16617524 8343963 4294184 ; leg controller genes
             )
            0)))
         (locomotion-controller ; Fetch the locomotion controller from the cns
          (central-nervous-system-locomotion-controller cns)))
    ;; Adjust front femur-tibia neurons to genotype
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-l1 locomotion-controller))
     (genome-front-leg-chromosome genome)
     (locomotion-controller-l1 locomotion-controller))
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-r1 locomotion-controller))
     (genome-front-leg-chromosome genome)
     (locomotion-controller-r1 locomotion-controller))
    ;; Adjust middle femur-tibia neurons to genotype
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-l2 locomotion-controller))
     (genome-middle-leg-chromosome genome)
     (locomotion-controller-l2 locomotion-controller))
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-r2 locomotion-controller))
     (genome-middle-leg-chromosome genome)
     (locomotion-controller-r2 locomotion-controller))
    ;; Adjust back femur-tibia neurons to genotype
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-l3 locomotion-controller))
     (genome-back-leg-chromosome genome)
     (locomotion-controller-l3 locomotion-controller))
    (adjust-to-chromosome! 
     (leg-controller-femur-tibia-neuron (locomotion-controller-r3 locomotion-controller))
     (genome-back-leg-chromosome genome)
     (locomotion-controller-r3 locomotion-controller))
    cns))

(defparameter *num-integrations-per-physical-updates* 3)

(defun fitness! (genome &optional (num-breve-synchs 200))
  "Destructively update the fitness of the chromosomes in the genome."
  (breve-interface:execute-breve-command "reset")
  (sleep 0.5) ; wait for creature to settle
  (let ((cns (genotype->phenotype genome))
        (energy 0)
        (start-location (get-location))
        (start-time (read-from-string (execute-breve-command-with-result "get-time"))))
    (dotimes (var num-breve-synchs)
      (dotimes (var *num-integrations-per-physical-updates*)
        (incf energy
              (synchronously-update-membrane-potentials!
               (central-nervous-system-locomotion-controller cns))))
      (synchronize-with-breve (central-nervous-system-breve-object cns)))
    (let ((stop-time (read-from-string (execute-breve-command-with-result "get-time"))))
      (format t " Average discrepancy between sim. and net. = ~As~%"
              (/ (abs (- (* *num-integrations-per-physical-updates* num-breve-synchs ctrnn::*timestep*)
                         (- stop-time start-time))) num-breve-synchs))
      (setf *num-integrations-per-physical-updates* (round (* (/ (- stop-time start-time) num-breve-synchs) 100)))
      (format t " (New) Num integrations per breve update = ~A~%" *num-integrations-per-physical-updates*))
    (let* ((distance-walked (- (get-location) start-location))
           (fitness (/ energy distance-walked))) ; Energy ?
      (format t " Fitness of this creature: ~A~%" fitness)
      ;; Update chromosomes
      (incf (chromosome-cumulative-fitness (genome-front-leg-chromosome genome)) fitness)
      (incf (chromosome-num-fitness-evaluations (genome-front-leg-chromosome genome)))
      (incf (chromosome-cumulative-fitness (genome-middle-leg-chromosome genome)) fitness)
      (incf (chromosome-num-fitness-evaluations (genome-middle-leg-chromosome genome)))
      (incf (chromosome-cumulative-fitness (genome-back-leg-chromosome genome)) fitness)
      (incf (chromosome-num-fitness-evaluations (genome-back-leg-chromosome genome)))))
  genome)

(defun sort-population (population)
  "Create a sorted (by fitness) version of 'population'."
  (labels ((chromosome-fitness (chromosome)
             (if (not (= (chromosome-num-fitness-evaluations chromosome) 0))
               (/ (chromosome-cumulative-fitness chromosome) (chromosome-num-fitness-evaluations chromosome))
               0)))
    (sort (copy-list population)
          #'(lambda (one two)
              (> (chromosome-fitness one)
                 (chromosome-fitness two))))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun evaluate! (front middle back &optional (num-breve-synchs 200))
  "Update fitness of neurons in each population."
  ;; Cross-evaluate chromosomes by iterating through each population and
  ;; realizing a phenotype by picking two random neurons from the other
  ;; populations. In theory there is a possibility (though not a very large probability)
  ;; that every chromosome is evaluated twice per generation using this strategy.
  (loop for chromosome in front
        do (let ((m (nth (random (length middle)) middle))
                 (b (nth (random (length back)) back)))
             (format t "Evaluating ~A~% ~A~% ~A~%" chromosome m b)
             (fitness! (make-genome :front-leg-chromosome chromosome
                                    :middle-leg-chromosome m
                                    :back-leg-chromosome b)
                       num-breve-synchs)))
    (loop for chromosome in middle
        do (let ((f (nth (random (length front)) front))
                 (b (nth (random (length back)) back)))
             (format t "Evaluating ~A~% ~A~% ~A~%" f chromosome b)
             (fitness! (make-genome :front-leg-chromosome f
                                    :middle-leg-chromosome chromosome
                                    :back-leg-chromosome b)
                       num-breve-synchs)))
      (loop for chromosome in back
        do (let ((m (nth (random (length middle)) middle))
                 (f (nth (random (length front)) front)))
             (format t "Evaluating ~A~% ~A~% ~A~%" f m chromosome)
             (fitness! (make-genome :front-leg-chromosome f
                                    :middle-leg-chromosome m
                                    :back-leg-chromosome chromosome)
                       num-breve-synchs))))

(defun recombine (chromosomes)
  "Create two children from a list of two chromosomes, using one/two-point crossover."
  ;; Assuming that in the initial population, some neurons has connections to the
  ;; right neurons, and other have proper weights. From this assumption, the
  ;; search for successful genome must find the gene combination that reflect both these.
  (let* ((p-1 (first chromosomes))
         (p-2 (second chromosomes))
         (c-1 (make-chromosome :bias-gene (chromosome-bias-gene p-1)
                               :time-constant-gene (chromosome-time-constant-gene p-1)
                               :dendrites-genes (copy-list (chromosome-dendrites-genes p-1))))
         (c-2 (make-chromosome :bias-gene (chromosome-bias-gene p-2)
                               :time-constant-gene (chromosome-time-constant-gene p-2)
                               :dendrites-genes (copy-list (chromosome-dendrites-genes p-2)))))
    ;; Swap bias-gene in 50\% of the cases
;    (when (< (random 100) 50)
;      (setf (chromosome-bias-gene c-1) (chromosome-bias-gene p-2)
;            (chromosome-bias-gene c-2) (chromosome-bias-gene p-1)))
    ;; Same goes for time-constant
;    (when (< (random 100) 50)
;      (setf (chromosome-time-constant-gene c-1) (chromosome-bias-gene p-2)
;            (chromosome-time-constant-gene c-2) (chromosome-time-constant-gene p-1)))
    ;; Always interchange one of the weights
    (let ((x-point (random (min (length (chromosome-dendrites-genes p-1))
                                (length (chromosome-dendrites-genes p-2))))))
      (setf (cdr (nth x-point (chromosome-dendrites-genes p-1)))
            (cdr (nth x-point (chromosome-dendrites-genes p-2))))
      (setf (cdr (nth x-point (chromosome-dendrites-genes p-2)))
            (cdr (nth x-point (chromosome-dendrites-genes p-1)))))
    (list c-1 c-2)))
             
         
(defun mutate! (chromosome)
  (let ((random-1 (random 3)))
    (cond ((= random-1 0) ; tamper with bias
           (incf (chromosome-bias-gene chromosome) (- (random 2000) 1000)))
          ((= random-1 1) ; tamper with time constant
           (incf (chromosome-time-constant-gene chromosome) (- (random 2000) 1000)))
          ((and (= random-1 2) ; tamper with one of the weights
                (> (length (chromosome-dendrites-genes chromosome)) 1))
           (let ((random-2 (random (length (chromosome-dendrites-genes chromosome))))) ;; Which weight to tamper with
             (incf (cdr (nth random-2 (chromosome-dendrites-genes chromosome))) (- (random 2000) 1000)))))))

(defun sane-esp (num-generations)
  (breve-interface:execute-breve-command "set-output-filter" -1)
  (let* (;(carausius-morosus::*maximum-leg-force* 60000)
         (carausius-morosus::*maximum-leg-down-force* 80000)
         (network-length (length (neural-network-neurons (make-instance 'carausius-morosus::leg-controller))))
         (population-size 16) 
         ;; One population per neuron type
         (front-neuron-population (make-population population-size network-length))
         (middle-neuron-population (make-population population-size network-length))
         (back-neuron-population (make-population population-size network-length))
         (plot-pane (make-instance 'capi-extensions:plot-pane
                                  :x-min -2 :x-max num-generations :x-axis (/ num-generations 10)
                                  :y-min (- (expt 10 10))
                                  :y-max (expt 10 11)
                                  :y-axis (expt 10 10)))
         (front-graph (make-instance 'capi-extensions:plot-pane-graph :colour :blue4))
         (middle-graph (make-instance 'capi-extensions:plot-pane-graph :colour :red4))
         (back-graph (make-instance 'capi-extensions:plot-pane-graph :colour :green4))
         (front-graph-avg (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
         (middle-graph-avg (make-instance 'capi-extensions:plot-pane-graph :colour :red))
         (back-graph-avg (make-instance 'capi-extensions::plot-pane-graph :colour :green)))
    ;; Evaluate all
    (evaluate! front-neuron-population middle-neuron-population back-neuron-population 500)
    ;; Sort population
    (setf front-neuron-population (sort-population front-neuron-population))
    (setf middle-neuron-population (sort-population middle-neuron-population))
    (setf back-neuron-population  (sort-population back-neuron-population))
    (push front-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push middle-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push back-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push front-graph-avg (capi-extensions:plot-pane-graphs plot-pane))
    (push middle-graph-avg (capi-extensions:plot-pane-graphs plot-pane))
    (push back-graph-avg (capi-extensions:plot-pane-graphs plot-pane))
    (capi:contain plot-pane)
    (do ((generation 1 (+ generation 1)))
        ((> generation num-generations))
      (format t "*** Generation ~A~%" generation)
      ;; Sort populations
      (setf front-neuron-population (sort-population front-neuron-population))
      (setf middle-neuron-population (sort-population middle-neuron-population))
      (setf back-neuron-population (sort-population back-neuron-population))
      ;; Recombine best two
      (let ((front-parents (select-parents front-neuron-population 2))
            (middle-parents (select-parents middle-neuron-population 2))
            (back-parents (select-parents back-neuron-population 2)))
        (let ((front-children (recombine (list (first front-parents) (second front-parents))))
              (middle-children (recombine (list (first middle-parents) (second middle-parents))))
              (back-children (recombine (list (first back-parents) (second back-parents)))))
          ;; Maybe mutate front leg
          (if (< (random 100) 10)
            (progn
              (mutate! (first front-children))
              (mutate! (second front-children))))
          ;; Maybe mutate middle leg
          (if (< (random 100) 10)
            (progn
              (mutate! (first middle-children))
              (mutate! (second middle-children))))
          ;; Maybe mutate back leg
          (if (< (random 100) 10)
            (progn
              (mutate! (first back-children))
              (mutate! (second back-children))))
          ;; Replace
          (setf front-neuron-population (append (butlast front-neuron-population 2) front-children))
          (setf middle-neuron-population (append (butlast middle-neuron-population 2) middle-children))
          (setf back-neuron-population (append (butlast back-neuron-population 2) back-children))
          ;; Evaluate again
          (evaluate! front-neuron-population middle-neuron-population back-neuron-population 500))
        ;; Print stats
        (push (cons generation (chromosome-average-fitness (first front-neuron-population)))
              (capi-extensions::plot-pane-graph-points front-graph))
        (push (cons generation (chromosome-average-fitness (first middle-neuron-population)))
              (capi-extensions::plot-pane-graph-points middle-graph))
        (push (cons generation (chromosome-average-fitness (first back-neuron-population)))
              (capi-extensions::plot-pane-graph-points back-graph))
        (push (cons generation (/ (loop for chromosome in front-neuron-population
                                        sum (chromosome-average-fitness chromosome))
                                  population-size))
              (capi-extensions::plot-pane-graph-points front-graph-avg))
        (push (cons generation (/ (loop for chromosome in middle-neuron-population
                                        sum (chromosome-average-fitness chromosome))
                                  population-size))
              (capi-extensions::plot-pane-graph-points middle-graph-avg))
        (push (cons generation (/ (loop for chromosome in back-neuron-population
                                        sum (chromosome-average-fitness chromosome))
                                  population-size))
              (capi-extensions::plot-pane-graph-points back-graph-avg))
        (gp:invalidate-rectangle plot-pane)))
    (list (first front-neuron-population) (first middle-neuron-population) (first back-neuron-population))))
      
      