;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Applying the Genetic Algorithm to the Beer-part of the leg controller
;;;;
;;;; * What is given:
;;;;
;;;; 0 1 2 3:
;;;; - Pacemaker unit neurons (different for each neuron):
;;;;   - Bias [-20 20]/[-20 0]
;;;;   - Time constant [ctrnn:*timestep* 10]
;;;;
;;;; 4 5 6 7:
;;;; - Weights between pacemaker unit neurons:
;;;;   - P1 -> P1 [4 20]
;;;;   - P1 -> P2 [-20 20]
;;;;   - P2 -> P1 [4 20]
;;;;   - P2 -> P2 [-20 20]
;;;;
;;;; 8 9:
;;;; - Stance neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 16]
;;;;
;;;; 10 11:
;;;; - Swing neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 16]
;;;; 
;;;; 12 13:
;;;; - Foot neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep* 16]
;;;;
;;;; 14 15:
;;;; - Sensor neurons (all are equal):
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep 16] 
;;;;
;;;; 16 17 18 19 20 21 22 23 24: 
;;;; - Weights between neurons:
;;;;   - pacemaker->stance [-20 0]
;;;;   - pacemaker->swing [0 20]
;;;;   - pacemaker->foot [-20 0]
;;;;   - backwd.s.->pacemaker [0 20]
;;;;   - forwrd.s.->pacemaker [-20 0]
;;;;   - forwrd.s.->stance [0 20]
;;;;   - forwrd.s.->swing [-20 0]
;;;;   - forwrd.s.->foot [0 20]
;;;;   - between all pacemakers [-20 0]
;;;;
;;;; What to evolve:
;;;;
;;;; 25 26:
;;;; - Bias and time-constants for back femur-tibia neurons:
;;;;   Biases [-20 20]
;;;;   Time constants [ctrnn:*timestep 16]
;;;;
;;;; 27:
;;;; - Weights from forward swing to forward femur-tibia neurons:
;;;;   [-20 20]
;;;;
;;;; 28:
;;;; - Weight from forward stance neuron to forward femur-tibia neuron:
;;;;   [-20 20]
;;;;
;;;; 29 30:
;;;; - Bias and time-constants for front femur-tibia neurons
;;;;   Biases [-20 20]
;;;;   Time constants [ctrnn:*timestep 16]
;;;;
;;;; 31:
;;;; - Weights from back swing neurons to back femur-tibia neurons 
;;;;   [-20 20]
;;;;
;;;; 32:
;;;; - Weight from back stance neuron to back femur-tibia neuron:
;;;;   [-20 20]
;;;;
;;;; Genotype representation is a cons on the form:
;;;;   ('vector of unsigned bytes' . fitness)
;;;;
;;;; This GA corresponds to the creature in carausius-morosus-3
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(in-package :leg-controller)

(setf carausius-morosus::*AEP-L1* (+ (* 1.1 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.1 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.55 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.55 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.5 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.5 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.3 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.3 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.45 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.45 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 3.3 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 3.3 (/ pi 4)) 0.01))

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; GA functions
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-population (size)
  "Create a list of length 'size' of random genotypes."
  (loop for n from 1 to size
        collect (let ((a (make-array 33 :element-type '(unsigned-byte 24)))) ; about 16,8 mil. values
                  ;; Set to random values
                  (loop for n from 0 to 32
                        do (cond
                            ((or (= n 27) (= n 32))
                             (setf (aref a n)
                                   (make-random-number-between 0 (/ (expt 2 24) 2))))
                            ((or (= n 28) (= n 31))
                             (setf (aref a n)
                                   (make-random-number-between (/ (expt 2 24) 2) (expt 2 24))))
                            (t
                             (setf (aref a n) (make-random-number-between 0 (expt 2 24))))))
                  (loop for n from 0 to 32
                        do ; Adjust biases to weights
                        (cond ((or (= n 25) (= n 29))
                               (setf (aref a n)
                                     (let ((weight-sum
                                            (+ (+ (* (- (aref a (+ n 2)) (expt 2 24)) (/ 40 (expt 2 24))) 20)
                                               (+ (* (- (aref a (+ n 3)) (expt 2 24)) (/ 40 (expt 2 24))) 20))))
                                       (round ; Must be represented by unsiged-byte
                                        (+ (/ (- (/ (- weight-sum) 2) 20) (/ 40 (expt 2 24))) (expt 2 24))))))))
                  ;; Replace the first 8 bytes by values evolved from pacemaker evolution
                  (setf (subseq a 0 8)
                        (vector 8741825 343682 3759221 2539167 6218767 2694810 7392821 16234160)) ; pacemaker genes
                  ;; Replace the next 17 by values evolved from leg controller take-1
                  (setf (subseq a 8 25)
                        (vector 8388608 13983186 9437184 4974807 7561010 3910349 8388608 5852116 8367905
                                8332011 4209360 8331748 14925040 8442354 12704632 8295991 4193451))
                  (cons a 0)))) ; Individual initialized neither fit nor unfit

(defun genotype->phenotype (genotype)
  "Convert a genotype into a creature that is suitable for fitness evaluation."
  (let* ((cns (make-instance 'central-nervous-system))
         (genome (car genotype))
         (locomotion-controller
          (central-nervous-system-locomotion-controller cns))
         (leg-controllers
          (list (locomotion-controller-l1 locomotion-controller)
                (locomotion-controller-l2 locomotion-controller)
                (locomotion-controller-l3 locomotion-controller)
                (locomotion-controller-r1 locomotion-controller)
                (locomotion-controller-r2 locomotion-controller)
                (locomotion-controller-r3 locomotion-controller))))
    ;; Set pacemaker neuron variables from the genome
    (dolist (l leg-controllers)
      (setf (neuron-bias (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (+ (* (- (aref genome 0) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (- (* (+ (aref genome 1) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*)) (setf (neuron-bias (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
            (+ (* (- (aref genome 2) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
            (- (* (+ (aref genome 3) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Create synapses within the pacemaker circuit
      ;; IO-neuron
      (setf (neuron-dendrites (pacemaker-io-neuron (leg-controller-pacemaker l))) nil)
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
                      ;; Self conn. gene into a value 4 < gene > 16
                      :strength (+ 4 (/ (aref genome 4) (/ (expt 2 24) 12)))))
      (add-synapse! 
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (pacemaker-hidden-neuron (leg-controller-pacemaker l))
                      ;; Other conn. into a value -20 < gene > 20
                      :strength (+ -20 (/ (aref genome 5) (/ (expt 2 24) 40)))))
      ;; H-neuron
      (setf (neuron-dendrites (pacemaker-hidden-neuron (leg-controller-pacemaker l))) nil)
      (add-synapse!
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse 
                      :from-neuron (pacemaker-hidden-neuron (leg-controller-pacemaker l))
                      :strength (+ 4 (/ (aref genome 6) (/ (expt 2 24) 12)))))
      (add-synapse! 
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
                      :strength  (+ -20 (/ (aref genome 7) (/ (expt 2 24) 40)))))
      ;; Set stance and swing neurons bias/time
      (setf (neuron-bias (leg-controller-stance-neuron l)) ;; [-20 20]
            (+ (* (- (aref genome 8) (expt 2 24)) (/ 40 (expt 2 24))) 20)) 
      (setf (neuron-time-constant (leg-controller-stance-neuron l)) ;; [0 16] + timestep 
            (- (* (+ (aref genome 9) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (setf (neuron-bias (leg-controller-swing-neuron l)) ;; [-20 20]
            (+ (* (- (aref genome 10) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-swing-neuron l)) ;; [0 16] + timestep 
            (- (* (+ (aref genome 11) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Foot neuron bias/time
      (setf (neuron-bias (leg-controller-foot-neuron l)) ;; [-20 20]
            (+ (* (- (aref genome 12) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-foot-neuron l)) ;; [0 16] + timestep 
            (- (* (+ (aref genome 13) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Sensor neurons (forward back)
      (setf (neuron-bias (leg-controller-forward-angle-sensor l)) ;; [-20 20]
            (+ (* (- (aref genome 14) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-forward-angle-sensor l)) ;; [0 16] + timestep 
            (- (* (+ (aref genome 15) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (setf (neuron-bias (leg-controller-backward-angle-sensor l)) ;; [-20 20]
            (+ (* (- (aref genome 14) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-backward-angle-sensor l)) ;; [0 16] + timestep 
            (- (* (+ (aref genome 15) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Weights
      (add-synapse!
       (leg-controller-stance-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
                      :strength (* (- (aref genome 16) (expt 2 24)) (/ 20 (expt 2 24))))) ; [- 20 0]
      (add-synapse!
       (leg-controller-swing-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
                      :strength (* (aref genome 17) (/ 20 (expt 2 24))))) ; [0 20]
      (add-synapse!
       (leg-controller-foot-neuron l)
       (make-instance 'synapse
                      :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
                      :strength (* (- (aref genome 18) (expt 2 24)) (/ 20 (expt 2 24))))) ; [- 20 0]
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (leg-controller-backward-angle-sensor l)
                      :strength (* (aref genome 19) (/ 20 (expt 2 24))))) ; [0 20]
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance 'synapse
                      :from-neuron (leg-controller-forward-angle-sensor l)
                      :strength (* (- (aref genome 20) (expt 2 24)) (/ 20 (expt 2 24))))) ; [- 20 0]
      (add-synapse!
       (leg-controller-stance-neuron l)
       (make-instance 'synapse
                      :from-neuron (leg-controller-forward-angle-sensor l)
                      :strength (* (aref genome 21) (/ 20 (expt 2 24))))) ; [0 20]
      (add-synapse!
       (leg-controller-swing-neuron l)
       (make-instance 'synapse
                      :from-neuron (leg-controller-forward-angle-sensor l)
                      :strength (* (- (aref genome 22) (expt 2 24)) (/ 20 (expt 2 24))))) ; [- 20 0]
      (add-synapse!
       (leg-controller-foot-neuron l)
       (make-instance 'synapse
                      :from-neuron (leg-controller-forward-angle-sensor l)
                      :strength (* (aref genome 23) (/ 20 (expt 2 24))))) ; [0 20]
      (carausius-morosus::set-up-connections locomotion-controller
                                             (* (- (aref genome 24) (expt 2 24)) (/ 20 (expt 2 24))))) ; [- 20 0]
    ;; Take 2
    ;; Adjust front neurons
    (let ((leg (locomotion-controller-l1 locomotion-controller)))
      (setf (neuron-bias (leg-controller-femur-tibia-neuron leg))
            (+ (* (- (aref genome 25) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-femur-tibia-neuron leg))
            (- (* (+ (aref genome 26) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-swing-neuron leg)
                      :strength (+ (* (- (aref genome 27) (expt 2 24)) (/ 40 (expt 2 24))) 20)))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-stance-neuron leg)
                      :strength (+ (* (- (aref genome 28) (expt 2 24)) (/ 40 (expt 2 24))) 20))))
    (let ((leg (locomotion-controller-r1 locomotion-controller)))
      (setf (neuron-bias (leg-controller-femur-tibia-neuron leg))
            (+ (* (- (aref genome 25) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-femur-tibia-neuron leg))
            (- (* (+ (aref genome 26) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-swing-neuron leg)
                      :strength (+ (* (- (aref genome 27) (expt 2 24)) (/ 40 (expt 2 24))) 20)))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-stance-neuron leg)
                      :strength (+ (* (- (aref genome 28) (expt 2 24)) (/ 40 (expt 2 24))) 20))))
    ;; Adjust back femur-tibia neurons
    (let ((leg (locomotion-controller-l3 locomotion-controller)))
      (setf (neuron-bias (leg-controller-femur-tibia-neuron leg))
            (+ (* (- (aref genome 29) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-femur-tibia-neuron leg))
            (- (* (+ (aref genome 30) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-swing-neuron leg)
                      :strength (+ (* (- (aref genome 31) (expt 2 24)) (/ 40 (expt 2 24))) 20)))
      (add-synapse!
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-stance-neuron leg)
                      :strength (+ (* (- (aref genome 32) (expt 2 24)) (/ 40 (expt 2 24))) 20))))
    (let ((leg (locomotion-controller-r3 locomotion-controller)))
      (setf (neuron-bias (leg-controller-femur-tibia-neuron leg))
            (+ (* (- (aref genome 29) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-femur-tibia-neuron leg))
            (- (* (+ (aref genome 30) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-swing-neuron leg)
                      :strength (+ (* (- (aref genome 31) (expt 2 24)) (/ 40 (expt 2 24))) 20)))
      (add-synapse! 
       (leg-controller-femur-tibia-neuron leg)
       (make-instance 'synapse
                      :from-neuron (leg-controller-stance-neuron leg)
                      :strength (+ (* (- (aref genome 32) (expt 2 24)) (/ 40 (expt 2 24))) 20))))
    cns))

(defparameter *num-integrations-per-physical-updates* 3)

(defun fitness (genotype &optional (num-breve-synchs 200))
  "Evaluate the fitness of a genotype, and return the fitness."
  (breve-interface:execute-breve-command "reset")
  (sleep 1.0)
  (let ((cns (genotype->phenotype genotype))
        (energy 0)
        (start-location (get-location))
        (start-time (read-from-string (execute-breve-command-with-result "get-time"))))
    (dotimes (var num-breve-synchs)
      (dotimes (var *num-integrations-per-physical-updates*)
        (incf energy
              (synchronously-update-membrane-potentials!
               (central-nervous-system-locomotion-controller cns))))
      (synchronize-with-breve (central-nervous-system-breve-object cns)))
    (let ((stop-time (read-from-string (execute-breve-command-with-result "get-time"))))
      (format t " Average discrepancy between sim. and net. = ~As~%"
              (/ (abs (- (* *num-integrations-per-physical-updates* num-breve-synchs ctrnn::*timestep*)
                         (- stop-time start-time))) num-breve-synchs))
      (setf *num-integrations-per-physical-updates* (round (* (/ (- stop-time start-time) num-breve-synchs) 100)))
      (format t " (New) Num integrations per breve update = ~A~%" *num-integrations-per-physical-updates*))
    (let ((distance-walked (- (get-location) start-location)))
      (/ distance-walked 1)))) ; NOTE: energy
  
(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
                                   (/ (+ (fitness genotype num-breve-synchs)
                                         (fitness genotype num-breve-synchs))
                                      2))))
                                         
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 30 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 25 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (random 2))
        (rand-nr (random 1000000)))
    (if (= rand 1)
      (if (< (- (expt 2 24) rand-nr) (aref new-genome x))
        (incf (aref new-genome x) rand-nr)
        (decf (aref new-genome x) rand-nr))
      (if (> rand-nr (aref new-genome x))
        (decf (aref new-genome x) rand-nr)
        (incf (aref new-genome x) rand-nr)))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations)
  (breve-interface:execute-breve-command "set-output-filter" -1)
  (let* ((carausius-morosus::*maximum-femur-tibia-force* 4000)
         (carausius-morosus::*maximum-leg-force* 40000)
         (carausius-morosus::*maximum-leg-down-force* 40000) ;; NOTE: 70000 = "minimum" to stand on three legs
         #+Lispworks(plot-pane (make-instance 'capi-extensions:plot-pane
                                  :x-min -2 :x-max num-generations :x-axis (/ num-generations 10)
                                  :y-min -10 ; (- (expt 10 6)) ;-10 ;-0.000000001
                                  :y-max 150 ; (* (expt 10 7) 5) ;100 ; 0.00000002
                                  :y-axis 10)); (expt 10 6))) ; 10)); 0.000000002))
	 #+Lispworks(best-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
	 #+Lispworks(avg-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :yellow))
	 #+Lispworks(worst-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :red))
        (population (sort-population (evaluate (make-random-population 50) 500))))
    #+Lispworks
    (progn
      (push best-graph (capi-extensions:plot-pane-graphs plot-pane))
      (push avg-graph (capi-extensions:plot-pane-graphs plot-pane))
      (push worst-graph (capi-extensions:plot-pane-graphs plot-pane))
      (capi:contain plot-pane))
    (do ((generation 1 (+ generation 1)))
        ((> generation num-generations))
      (format t "*** Generation ~A~%" generation)
      ;; Select parents
      (let ((parents (select-parents population 2)))
        ;; Recombine
        (let ((children (recombine parents)))
          ;; Maybe mutate
          (when (< (random 100) 5)
            (setf children (loop for c in children collect (mutate c))))
          ;; Evaluate new individuals
          (setf children (evaluate children 500))
          ;; Select individuals for next generation
          (setf population (sort-population (append
                                             children
                                             (remove-individuals population 2))))
          ;; Update display
          #+Lispworks
	  (progn
	    (push (cons generation (cdr (first population)))
		  (capi-extensions::plot-pane-graph-points best-graph))
	    (push (cons generation (/ (loop for ind in population
					 sum (cdr ind))
				      (length population)))
		  (capi-extensions::plot-pane-graph-points avg-graph))
	    (push (cons generation (cdr (first (last population))))
		  (capi-extensions::plot-pane-graph-points worst-graph))
	    (gp:invalidate-rectangle plot-pane)))))
    (first population)))

(defun test-run ()
  (let ((carausius-morosus::*maximum-leg-force* 40000)
        (carausius-morosus::*maximum-leg-down-force* 40000))
    (breve-interface:execute-breve-command "start-recording-movie")
    (fitness (Cons (vector
                    8741825 343682 3759221 2539167 6218767 2694810 7392821 16878403 8388608 13983186 9437184 4974807 7561010 3910349 8388608 5852116 8367905 8332011 4209360 8331748 14925040 8442354 12704632 8295991 4193451 7924026 6860218 7753033 9953348 8322310 3306156 12938182 3971630)
                   0)
             5000)))
