;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Applying the Genetic Algorithm to the Beer-part of the leg controller
;;;;
;;;; * What is given:
;;;;
;;;; 0 1 2 3:
;;;; - Pacemaker unit neurons (different for each neuron):
;;;;   - Bias [-20 20]/[-20 0]
;;;;   - Time constant [ctrnn:*timestep* 10]
;;;;
;;;; 4 5 6 7:
;;;; - Weights between pacemaker unit neurons:
;;;;   - P1 -> P1 [4 20]
;;;;   - P1 -> P2 [-20 20]
;;;;   - P2 -> P1 [4 20]
;;;;   - P2 -> P2 [-20 20]
;;;;
;;;; 8 9:
;;;; - Stance neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;;
;;;; 10 11:
;;;; - Swing neurons:
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;; 
;;;; 12 13:
;;;; - Foot neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep* 10]
;;;;
;;;; 14 15:
;;;; - Sensor neurons (all are equal):
;;;;   - External current magnitude [0 255]
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep 10] 
;;;;
;;;; * What to evolve:
;;;;
;;;; 16 17:
;;;; - Back leg femur-tibia neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;;
;;;; 18 19:
;;;; - Middle leg femur-tibia neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;; 20 21:
;;;; - Front leg femur-tibia neuron
;;;;   - Bias [-20 20]
;;;;   - Time constant [*ctrnn:*timestep* 10]
;;;;
;;;; 22 23 24:
;;;; - Input weight to each neuron from pacemaker or stance [-20 20]
;;;;
;;;; 25 26 27:
;;;; - Input weight to each neuron from forward sensor or swing neuron [-20 20]
;;;;
;;;; 28:
;;;; - Gene deciding what neurons the new neurons should connect to
;;;;
;;;; Genotype representation is a cons on the form:
;;;;   ('vector of unsigned bytes' . fitness)
;;;;
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(in-package :leg-controller)


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Utility functions, unrelated to the GA
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-number-between (min max &optional (decimals 0))
  "Create a random number between min and max (inclusive)."
  (let ((max (* (expt 10 decimals) max))
        (min (* (expt 10 decimals) min)))
    (/ (+ min (random (- max min -1))) (expt 10 decimals))))

(defun get-location ()
  "Get the location of the breve creature."
  (car
   (read-from-string
    (coerce
     (loop for char across
           (breve-interface:execute-breve-command-with-result "get-location")
           unless (char-equal #\, char)
           collect char) 'string))))


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; GA functions
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-population (size)
  "Create a list of length 'size' of random genotypes."
  (loop for n from 1 to size
        collect (let ((a (make-array 29 :element-type '(unsigned-byte 17)))) ; about 16,8 mil. values
                  (loop for n from 0 to 28
                        do (setf (aref a n) (make-random-number-between 0 (expt 2 24))))
                  ;; Replace the first 8 bytes by values evolved from pacemaker evolution
                  (setf (subseq a 0 8)
                        (vector 8741825 343682 3759221 2539167 6218767 2694810 7392821 16234160))
                  ;; Replace the next 9 bytes by values evolved from locomotion controller evolution
                  (setf (subseq a 8 15)
                        (vector 8388608 12515281 9437184 16472653 7521170 289609 8388608 739603))
                  (cons a 0)))) ; Individual initialized neither fit nor unfit

(defun genotype->phenotype (genotype)
  "Convert a genotype into a creature that is suitable for fitness evaluation."
  (let* ((cns (make-instance 'central-nervous-system))
         (genome (car genotype))
         (locomotion-controller
          (central-nervous-system-locomotion-controller cns))
         (leg-controllers
          (list (locomotion-controller-l1 locomotion-controller)
                (locomotion-controller-l2 locomotion-controller)
                (locomotion-controller-l3 locomotion-controller)
                (locomotion-controller-r1 locomotion-controller)
                (locomotion-controller-r2 locomotion-controller)
                (locomotion-controller-r3 locomotion-controller))))
    ;; NOTE: Inhibit the LC-neuron for now
    (setf (neuron-external-current
           (locomotion-controller-lc-neuron
            (central-nervous-system-locomotion-controller cns)))
          -20)
    ;; Set pacemaker neuron variables from the genome
    (dolist (l leg-controllers)
      (setf (neuron-bias
             (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (+ (* (- (aref genome 0) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant
             (pacemaker-io-neuron (leg-controller-pacemaker l)))
            (- (* (+ (aref genome 1) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (setf (neuron-bias (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
            (+ (* (- (aref genome 2) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant
             (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
            (- (* (+ (aref genome 3) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Create synapses within the pacemaker circuit
      ;; IO-neuron
      (setf
       (neuron-dendrites (pacemaker-io-neuron (leg-controller-pacemaker l)))
       nil)
      (add-synapse!
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance
        'synapse
        :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
        ;; Self conn. gene into a value 4 < gene > 16
        :strength (+ 4 (/ (aref genome 4) (/ (expt 2 24) 12)))))
      (add-synapse! 
       (pacemaker-io-neuron (leg-controller-pacemaker l))
       (make-instance
        'synapse
        :from-neuron (pacemaker-hidden-neuron (leg-controller-pacemaker l))
        ;; Other conn. into a value -20 < gene > 20
        :strength (+ -20 (/ (aref genome 5) (/ (expt 2 24) 40)))))
      ;; H-neuron
      (setf
       (neuron-dendrites (pacemaker-hidden-neuron (leg-controller-pacemaker l)))
       nil)
      (add-synapse!
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance
        'synapse 
        :from-neuron (pacemaker-hidden-neuron (leg-controller-pacemaker l))
        :strength (+ 4 (/ (aref genome 6) (/ (expt 2 24) 12)))))
      (add-synapse!
       (pacemaker-hidden-neuron (leg-controller-pacemaker l))
       (make-instance
        'synapse
        :from-neuron (pacemaker-io-neuron (leg-controller-pacemaker l))
        :strength  (+ -20 (/ (aref genome 7) (/ (expt 2 24) 40)))))
      ;; Set stance and swing neurons bias/time
      (setf (neuron-bias (leg-controller-stance-neuron l))
            ;; [-20 20]
            (+ (* (- (aref genome 8) (expt 2 24)) (/ 40 (expt 2 24))) 20)) 
      (setf (neuron-time-constant (leg-controller-stance-neuron l))
             ;; [0 16] + timestep 
            (- (* (+ (aref genome 9) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (setf (neuron-bias (leg-controller-swing-neuron l))
            ;; [-20 20]
            (+ (* (- (aref genome 10) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-swing-neuron l))
            ;; [0 16] + timestep 
            (- (* (+ (aref genome 11) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Foot neuron bias/time
      (setf (neuron-bias (leg-controller-foot-neuron l))
            ;; [-20 20]
            (+ (* (- (aref genome 12) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-foot-neuron l))
            ;; [0 16] + timestep 
            (- (* (+ (aref genome 13) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Sensor neurons (forward back)
      (setf (neuron-bias (leg-controller-forward-angle-sensor l))
            ;; [-20 20]
            (+ (* (- (aref genome 14) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-forward-angle-sensor l))
            ;; [0 16] + timestep 
            (- (* (+ (aref genome 15) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      (setf (neuron-bias (leg-controller-backward-angle-sensor l))
            ;; [-20 20]
            (+ (* (- (aref genome 14) (expt 2 24)) (/ 40 (expt 2 24))) 20))
      (setf (neuron-time-constant (leg-controller-backward-angle-sensor l))
            ;; [0 16] + timestep 
            (- (* (+ (aref genome 15) (expt 2 24)) (/ 16 (expt 2 24)))
               16 ctrnn::*timestep*))
      ;; Synaptic connections of femur-tibia joints
      (setf (neuron-dendrites (leg-controller-femur-tibia-neuron l)) nil)
      ;; From pacemaker or stance:
      (add-synapse!
       (leg-controller-femur-tibia-neuron l)
       (make-instance 'synapse
                      :from-neuron (if (> (aref genome 28) (/ (expt 2 24) 2))
                                     (pacemaker-io-neuron (leg-controller-pacemaker l))
                                     (leg-controller-stance-neuron l))
                      :strength (cond ((or (eq (locomotion-controller-l3 locomotion-controller) l)
                                           (eq (locomotion-controller-r3 locomotion-controller) l))
                                       (+ (* (- (aref genome 22) (expt 2 24)) (/ 40 (expt 2 24))) 20))
                                      ((or (eq (locomotion-controller-l2 locomotion-controller) l)
                                           (eq (locomotion-controller-r2 locomotion-controller) l))
                                       (+ (* (- (aref genome 23) (expt 2 24)) (/ 40 (expt 2 24))) 20))
                                      ((or (eq (locomotion-controller-l1 locomotion-controller) l)
                                           (eq (locomotion-controller-r1 locomotion-controller) l))
                                       (+ (* (- (aref genome 24) (expt 2 24)) (/ 40 (expt 2 24))) 20)))))
      ;; From sensor or swing
      (add-synapse!
       (leg-controller-femur-tibia-neuron l)
       (make-instance 'synapse
                      :from-neuron (if (> (aref genome 28) (/ (expt 2 24) 2))
                                     (leg-controller-forward-angle-sensor l)
                                     (leg-controller-swing-neuron l))
                      :strength (cond ((or (eq (locomotion-controller-l3 locomotion-controller) l)
                                           (eq (locomotion-controller-r3 locomotion-controller) l))
                                       (+ (* (- (aref genome 25) (expt 2 24)) (/ 40 (expt 2 24))) 20))
                                      ((or (eq (locomotion-controller-l2 locomotion-controller) l)
                                           (eq (locomotion-controller-r2 locomotion-controller) l))
                                       (+ (* (- (aref genome 26) (expt 2 24)) (/ 40 (expt 2 24))) 20))
                                      ((or (eq (locomotion-controller-l1 locomotion-controller) l)
                                           (eq (locomotion-controller-r1 locomotion-controller) l))
                                       (+ (* (- (aref genome 27) (expt 2 24)) (/ 40 (expt 2 24))) 20))))))
    ;; Back femur-tibia neurons
    (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-l3 locomotion-controller)))
          (+ (* (- (aref genome 16) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-l3 locomotion-controller)))
          (- (* (+ (aref genome 17) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
    (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-r3 locomotion-controller)))
          (+ (* (- (aref genome 16) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-r3 locomotion-controller)))
          (- (* (+ (aref genome 17) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
    ;; Middle femur-tibia neurons
    (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-l2 locomotion-controller)))
          (+ (* (- (aref genome 18) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-l2 locomotion-controller)))
          (- (* (+ (aref genome 19) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
        (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-r2 locomotion-controller)))
          (+ (* (- (aref genome 18) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-r2 locomotion-controller)))
          (- (* (+ (aref genome 19) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
    ;; Front femur-tibia neurons
    (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-l1 locomotion-controller)))
          (+ (* (- (aref genome 20) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-l1 locomotion-controller)))
          (- (* (+ (aref genome 21) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
    (setf (neuron-bias (leg-controller-femur-tibia-neuron (locomotion-controller-r1 locomotion-controller)))
          (+ (* (- (aref genome 20) (expt 2 24)) (/ 40 (expt 2 24))) 20))
    (setf (neuron-time-constant (leg-controller-femur-tibia-neuron (locomotion-controller-r1 locomotion-controller)))
          (- (* (+ (aref genome 21) (expt 2 24)) (/ 16 (expt 2 24)))
             16 ctrnn::*timestep*))
    
    cns))

(defparameter *num-integrations-per-physical-updates* 10)

(defun fitness (genotype &optional (num-breve-synchs 200))
  "Evaluate the fitness of a genotype, and return the fitness."
  (breve-interface:execute-breve-command "reset")
  (sleep 1.0)
  (let ((cns (genotype->phenotype genotype))
        (energy 0)
        (start-location (get-location))
        (start-time (read-from-string (execute-breve-command-with-result "get-time"))))
    (dotimes (var num-breve-synchs)
      (dotimes (var *num-integrations-per-physical-updates*)
        (incf energy
              (synchronously-update-membrane-potentials!
               (central-nervous-system-locomotion-controller cns))))
      (synchronize-with-breve (central-nervous-system-breve-object cns)))
    (let ((stop-time (read-from-string (execute-breve-command-with-result "get-time"))))
      (format t " Average discrepancy between sim. and net. = ~As~%"
              (/ (abs (- (* *num-integrations-per-physical-updates* num-breve-synchs ctrnn::*timestep*)
                         (- stop-time start-time))) num-breve-synchs))
      (setf *num-integrations-per-physical-updates* (round (* (/ (- stop-time start-time) num-breve-synchs) 100)))
      (format t " (New) Num integrations per breve update = ~A~%" *num-integrations-per-physical-updates*))
    (let ((distance-walked (- (get-location) start-location)))
      (/ distance-walked energy)))) ; NOTE: energy
  
(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
                                   (fitness genotype num-breve-synchs))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))
;  (let ((population (copy-list population)))
;    (loop for individual-num from 1 to num-to-slay
;          do (let ((individual
;                    (nth (make-random-number-between 
;                          (round (/ (length population) 2))
;                          (- (length population) 1))
;                         population)))
;               (setf population (remove individual population))))
;    population))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with one-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         ;(x 8)) ; split on pacemaker/rest
         (x (make-random-number-between 16 (- (length c1) 1))))
    (setf (subseq c1 0 x) (subseq (car p2) 0 x))
    (setf (subseq c2 0 x) (subseq (car p1) 0 x))
    (list (cons c1 0) (cons c2 0))))


(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 16 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (random 2))
        (rand-nr (random 10)))
    (if (= rand 1)
      (if (< (- (expt 2 24) rand-nr) (aref new-genome x))
        (incf (aref new-genome x) rand-nr)
        (decf (aref new-genome x) rand-nr))
      (if (> rand-nr (aref new-genome x))
        (decf (aref new-genome x) rand-nr)
        (incf (aref new-genome x) rand-nr)))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations)
  (breve-interface:execute-breve-command "set-output-filter" -1)
  (let* ((carausius-morosus::*maximum-leg-force* 5000)
         (carausius-morosus::*maximum-leg-down-force* 70000) ;; NOTE: 70000 = minimum to stand on three legs
         (plot-pane (make-instance 'capi-extensions:plot-pane
                                  :x-min -2 :x-max num-generations :x-axis (/ num-generations 10)
                                  :y-min -0.000001
                                  :y-max  0.00001
                                  :y-axis 0.000001))
        (best-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
        (avg-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :yellow))
        (worst-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :red))
        (population (sort-population (evaluate (make-random-population 50) 1000))))
    (push best-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push avg-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push worst-graph (capi-extensions:plot-pane-graphs plot-pane))
    (capi:contain plot-pane)
    (do ((generation 1 (+ generation 1)))
        ((> generation num-generations))
      (format t "*** Generation ~A~%" generation)
      ;; Select parents
      (let ((parents (select-parents population 2)))
        ;; Recombine
        (let ((children (recombine parents)))
          ;; Maybe mutate
          (when (< (random 100) 3)
            (setf children (loop for c in children collect (mutate c))))
          ;; Evaluate new individuals
          (setf children (evaluate children 1000))
          ;; Select individuals for next generation
          (setf population (sort-population (append
                                             (remove-individuals population 2)
                                             children)))
          ;; Update display
          (push (cons generation (cdr (first population)))
                (capi-extensions::plot-pane-graph-points best-graph))
          (push (cons generation (/ (loop for ind in population
                                          sum (cdr ind))
                                    (length population)))
                (capi-extensions::plot-pane-graph-points avg-graph))
          (push (cons generation (cdr (first (last population))))
                (capi-extensions::plot-pane-graph-points worst-graph))
          (gp:invalidate-rectangle plot-pane))))
    (first population)))
