;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; A pacemaker neural network consists of two interconnected neurons, where
;;;; each neuron has one self-connection.
;;;;
;;;; What to evolve:
;;;;
;;;; 0 1 2 3:
;;;;   - Bias [-20 20]
;;;;   - Time constant [ctrnn:*timestep* 16]
;;;;
;;;; 4 5 6 7:
;;;; - Weights between pacemaker unit neurons:
;;;;   - P1 -> P1 [4 20] (> 4 ensures a bifurcative neuron)
;;;;   - P1 -> P2 [-20 20]
;;;;   - P2 -> P1 [-20 20]
;;;;   - P2 -> P2 [4 20]
;;;;
;;;; Genotype representation is a cons on the form:
;;;;   ('vector of unsigned bytes' . fitness)
;;;;
;;;; NOTES:
;;;;
;;;; 1) The population is seeded with center-crossing neural networks to
;;;;    improve probability of oscillating circuits
;;;;    (see [Beer and Mathayomchan 2002].
;;;; 2) To start the simulation, run the pacemaker-world simulation, then start
;;;;    the GA (e.g. by evaluating "(ga 100)").
;;;;
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(in-package :pacemaker-evolution)


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; Utility functions, unrelated to the GA
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-number-between (min max &optional (decimals 0))
  "Create a random number between min and max (inclusive)."
  (let ((max (* (expt 10 decimals) max))
        (min (* (expt 10 decimals) min)))
    (/ (+ min (random (- max min -1))) (expt 10 decimals))))

(defun get-location ()
  "Shorthand to get the location of the breve creature."
  (car
   (read-from-string
    (coerce
     (loop for char across
           (breve-interface:execute-breve-command-with-result "get-location")
           unless (char-equal #\, char)
           collect char) 'string))))

(defun weight-gene->weight (gene)
  "Convert 24-bit integer into a real weight value."
  (+ -20 (/ gene (/ (expt 2 24) 40))))

(defun time-constant-gene->time-constant (gene)
  (+ (/ gene (/ (expt 2 24) 16)) ctrnn::*timestep*))

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;; GA functions
;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
(defun make-random-population (size)
  "Create a list of length 'size' of \"random\" genotypes."
  (loop
   for n from 1 to size
   collect (let ((a (make-array
                     8
                     :element-type '(unsigned-byte 24)))) ; about 16,8 mil. vals
             (loop for n from 0 to 7
                   do (if nil ;(or (= n 4) (= n 6))
                        (setf (aref a n) ; Self weights between 4 and ~16 = bif.
                              (make-random-number-between 
                               (round (/ (expt 2 24) 5/3)) (round (/ (expt 2 24) 9/8))))
                        (setf (aref a n) ; Otherwise random
                              (make-random-number-between 0 (expt 2 24)))))
             ;; Set low time-constants
             ;; Ensure that all networks created are center-crossing
             ;; by setting the biases to (/ (- (sum-of-weights)) 2)
             (setf (aref a 1) (make-random-number-between 450000 600000))
             (setf (aref a 3) (make-random-number-between 450000 600000))
;;;;              (setf (aref a 0) ; Bias of IO-neuron
;;;;                    (round ; Must be represented by unsiged-byte
;;;;                     (* (+ (/ (-
;;;;                               (+
;;;;                                ;; weight one for the neuron
;;;;                                (+ -20 (/ (aref a 4) (/ (expt 2 24) 40)))
;;;;                               ;; weight two for the neuron
;;;;                                (+ -20 (/ (aref a 5) (/ (expt 2 24) 40)))))
;;;;                              2)
;;;;                           ;; Add 20 ...
;;;;                           20)
;;;;                        ;; ... and multiply by this (= inverse of what is
;;;;                        ;; done in genotype->phenotype conversion) to get
;;;;                        ;; a value for the bias
;;;;                        (/ (expt 2 24) 40)))) 
;;;;              (setf (aref a 2) ; Bias of hidden neuron 
;;;;                    (round
;;;;                     (* (+ (/ (-
;;;;                               (+
;;;;                                (+ -20 (/ (aref a 6) (/ (expt 2 24) 40)))
;;;;                                (+ -20 (/ (aref a 7) (/ (expt 2 24) 40)))))
;;;;                              2)
;;;;                           20)
;;;;                    (/ (expt 2 24) 40))))
(cons a 0)))) ; Individual initialized neither fit nor unfit

(defun genotype->phenotype (genotype)
  "Convert a genotype into an individual suitable for fitness evaluation."
  (let* ((genome (car genotype))
         (pm (make-instance 'pacemaker)))
    ;; Set pacemaker neuron variables from the genome
    (setf (neuron-bias (pacemaker-io-neuron pm))
          (weight-gene->weight (aref genome 0)))
    (setf (neuron-time-constant (pacemaker-io-neuron pm))
          (time-constant-gene->time-constant (aref genome 1)))
    (setf (neuron-bias (pacemaker-hidden-neuron pm))
          (weight-gene->weight (aref genome 2)))
    (setf (neuron-time-constant (pacemaker-hidden-neuron pm))
          (time-constant-gene->time-constant (aref genome 3)))
    ;; Create synapses within the pacemaker circuit
    (setf (neuron-dendrites (pacemaker-io-neuron pm)) nil) ; Remove existing
    ;; IO neuron synapses:
    (add-synapse!
     (pacemaker-io-neuron pm)
     (make-instance 'synapse
                    :from-neuron (pacemaker-io-neuron pm)
                    :strength (weight-gene->weight (aref genome 4))))
    (add-synapse!
     (pacemaker-io-neuron pm)
     (make-instance 'synapse
                    :from-neuron (pacemaker-hidden-neuron pm)
                    :strength (weight-gene->weight (aref genome 5))))
    ;; Hidden neuron synapses (same procedure):
    (setf (neuron-dendrites (pacemaker-hidden-neuron pm)) nil)
    (add-synapse!
     (pacemaker-hidden-neuron pm)
     (make-instance 'synapse 
                    :from-neuron (pacemaker-hidden-neuron pm)
                    :strength (weight-gene->weight (aref genome 6))))
    (add-synapse!
     (pacemaker-hidden-neuron pm)
     (make-instance 'synapse
                    :from-neuron (pacemaker-io-neuron pm)
                    :strength (weight-gene->weight (aref genome 7))))
    ;; Return the resulting pacemaker neural network
    pm))

(defun fitness (genotype &optional (num-breve-synchs 200))
  "Evaluate the fitness of a genotype."
  (unless (gethash (car genotype) *fitnesses*)
    (setf (gethash (car genotype) *fitnesses*)
          (progn
            (breve-interface:execute-breve-command-with-result "reset")
            (loop ; Wait for creature to fall down on ground
             until (= 0 (read-from-string (breve-interface:execute-breve-command-with-result
                                           "get-integration-times")))
             do (sleep 0.01))
            (let ((pacemaker (genotype->phenotype genotype))
                  (start-location (get-location))
                  (start-time (read-from-string (breve-interface:execute-breve-command-with-result "get-time"))))
              ;; Start the integration loop
              (dotimes (var num-breve-synchs)
                ;; Update the neural net only
                (dotimes (var 5) ; Five updates per breve synch
                  (synchronously-update-membrane-potentials! pacemaker))
                ;; Synch with breve simulation
                (execute-breve-command-with-result ;; Set leg force depending on IO-neuron f. freq.
                                                   "set-leg-force"
                                                   (* 4 (- (ctrnn::neuron-snapshot-firing-frequency
                                                            (pacemaker-io-neuron pacemaker)) 0.5)))
                (loop
                 until (= 0 (read-from-string (breve-interface:execute-breve-command-with-result
                                               "get-integration-times")))
                 do (sleep 0.001)))
              (format t "time diff = ~A ~%" 
                      (- (- (read-from-string (breve-interface:execute-breve-command-with-result "get-time"))
                            start-time)
                         (* num-breve-synchs 5 ctrnn::*timestep*)))
              ;; Fitness is the distance walked by the creature (x(stop) - x(start))
              (let ((distance-walked (- (get-location) start-location)))
                distance-walked)))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and individual-num from 1 
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype) individual-num (length individuals))
                  (let ((res (cons (car genotype)
                                   (fitness genotype num-breve-synchs))))
                    (format t " Resulting fitness = ~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun select-parents (population &optional (num-parents 2))
  "Pick random (but not equal) parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine (two-point crossover) the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (x (make-random-number-between 0 (- (length c1) 2)))
         (y (make-random-number-between (+ x 1) (- (length c1) 1))))
    (setf (subseq c1 x y) (subseq (car p2) x y))
    (setf (subseq c2 x y) (subseq (car p1) x y))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 20000) 10000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (run-id 1))
  ;; Turn off the large amount of uninteresting console messages from breve
  (breve-interface:execute-breve-command "set-output-filter" -1000)
  (let (#+Lispworks(plot-pane
         (make-instance
          'capi-extensions:plot-pane
          :x-min -10 :x-max num-generations :x-axis (/ num-generations 10)
          :y-min -1 
          :y-max 20 
          :y-axis 5)) 
        #+Lispworks(best-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :blue))
        #+Lispworks(avg-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :yellow))
        #+Lispworks(worst-graph
         (make-instance 'capi-extensions:plot-pane-graph :colour :red))
	(ctrnn::*timestep* 0.01)
        (population
         (sort-population (evaluate (make-random-population 40) 160)))
        (log-file-name (format nil "logs/pacemaker-without-bifurcation-and-cc-log-~A.txt" run-id)))
    ;; Set up grap and display it
    #+Lispworks (progn
    (push best-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push avg-graph (capi-extensions:plot-pane-graphs plot-pane))
    (push worst-graph (capi-extensions:plot-pane-graphs plot-pane))
    (capi:contain plot-pane))
    (with-open-file
        (log-file log-file-name
                  :direction :output
                  :if-does-not-exist :create
                  :if-exists :supersede)
      ;; Run the GA
      (do ((generation 1 (+ generation 1)))
          ((> generation num-generations))
        (format t "*** Generation ~A~%" generation)
        ;; Select 4 parents
        (let ((parents (select-parents population 2)))
          ;; Recombine 1 & 3 and 2 & 4
          (let ((children (recombine (list (first parents) (second parents)))))
            ;; Maybe mutate the children
            (when (< (random 100) 5)
              (setf children (loop for c in children
                                   collect (mutate c))))
            ;; Evaluate new individuals
            (setf children (evaluate children 160))
            ;; Select individuals for next generation (replace worst 4)
            (setf population (sort-population (append (butlast population 2) children)))
            (format log-file "~A ~A ~A ~A ~%"
                    generation
                    (cdr (first population))
                    (/ (loop for ind in population
                             sum (cdr ind))
                       (length population))
                    (cdr (first (last population))))
            (force-output log-file)
            ;; Update graph display
            #+Lispworks(progn
                         (push (cons generation (cdr (first population)))
                               (capi-extensions::plot-pane-graph-points best-graph))
                         (push (cons generation (/ (loop for ind in population
                                                         sum (cdr ind))
                                                   (length population)))
                               (capi-extensions::plot-pane-graph-points avg-graph))
                         (push (cons generation (cdr (first (last population))))
                               (capi-extensions::plot-pane-graph-points worst-graph))
                         (gp:invalidate-rectangle plot-pane))))))
    ;; Return best genome
    (car (first population))))

(defun 15-runs ()
  (dotimes (time 15)
    (ga 150 time)))
